//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Instrument/CoordSystem2D.h
//! @brief     Defines interface CoordSystem2D and its subclasses.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_DEVICE_INSTRUMENT_COORDSYSTEM2D_H
#define BORNAGAIN_DEVICE_INSTRUMENT_COORDSYSTEM2D_H

#include "Base/Vector/Vectors3D.h"
#include "Device/Unit/ICoordSystem.h"

class Beam;
class IDetector;
class IDetector2D;
class Instrument;
class RectangularDetector;
class RectangularPixel;
class SphericalDetector;

//! Interface for objects that provide axis translations to different units for IDetector objects
//! @ingroup simulation_internal

class CoordSystem2D : public ICoordSystem {
public:
    CoordSystem2D(const Beam& beam);
    ~CoordSystem2D() override = default;

    virtual size_t dimension() const override { return m_axis_data_table.size(); }

    double calculateMin(size_t i_axis, Axes::Coords units) const override;
    double calculateMax(size_t i_axis, Axes::Coords units) const override;
    size_t axisSize(size_t i_axis) const override;

    //! Returns list of units that are available for all 2D detectors.
    //! Further units may be added by child classes.
    std::vector<Axes::Coords> availableUnits() const override;

    std::unique_ptr<const IAxis> createConvertedAxis(size_t i_axis,
                                                     Axes::Coords units) const override;

    //! Helper factory function to use in GISASSimulation. Depending on the type of detector,
    //! returns either RectangularConverter or SphericalConverter.
    static std::unique_ptr<CoordSystem2D> createConverterForGISAS(const Instrument& instrument);

protected:
    CoordSystem2D(const CoordSystem2D& other);
    void addDetectorAxis(const IDetector& detector, size_t i_axis);

    void addAxisData(std::string name, double min, double max, Axes::Coords default_units,
                     size_t nbins);

#ifndef SWIG
    struct AxisData {
        std::string name;
        double min, max;
        Axes::Coords default_units;
        size_t nbins;
    };
    std::vector<AxisData> m_axis_data_table;
#endif
    double m_wavelength;
    double m_alpha_i;
    double m_phi_i;

private:
    virtual double calculateValue(size_t i_axis, Axes::Coords units, double value) const = 0;
};


//! ICoordSystem class that handles the unit translations for spherical detectors
//! Its default units are radians for both axes
//! @ingroup simulation_internal

class SphericalConverter : public CoordSystem2D {
public:
    SphericalConverter() = delete;
    SphericalConverter(const SphericalDetector& detector, const Beam& beam);
    ~SphericalConverter() override;

    SphericalConverter* clone() const override;

    //! Returns the list of all available units
    std::vector<Axes::Coords> availableUnits() const override;

    Axes::Coords defaultUnits() const override { return Axes::Coords::DEGREES; }

private:
    SphericalConverter(const SphericalConverter& other); //!< used by clone()
    double calculateValue(size_t i_axis, Axes::Coords units, double value) const override;
    std::vector<std::map<Axes::Coords, std::string>> createNameMaps() const override;
};


//! ICoordSystem class that handles the unit translations for rectangular detectors
//! Its default units are mm for both axes
//! @ingroup simulation_internal

class RectangularConverter : public CoordSystem2D {
public:
    RectangularConverter() = delete;
    RectangularConverter(const RectangularDetector& detector, const Beam& beam);
    ~RectangularConverter() override;

    RectangularConverter* clone() const override;

    //! Returns the list of all available units
    std::vector<Axes::Coords> availableUnits() const override;

    Axes::Coords defaultUnits() const override { return Axes::Coords::MM; }

private:
    RectangularConverter(const RectangularConverter& other); //!< used by clone()
    double calculateValue(size_t i_axis, Axes::Coords units, double value) const override;
    std::vector<std::map<Axes::Coords, std::string>> createNameMaps() const override;

    std::unique_ptr<RectangularPixel> m_detector_pixel;
};


//! ICoordSystem class that handles the unit translations for off-specular simulations
//! with a spherical detector
//! Its default units are radians for both axes
//! @ingroup simulation_internal

class OffSpecularConverter : public CoordSystem2D {
public:
    OffSpecularConverter() = delete;
    OffSpecularConverter(const IDetector2D& detector, const Beam& beam, const IAxis& alpha_axis);
    ~OffSpecularConverter() override;

    OffSpecularConverter* clone() const override;

    Axes::Coords defaultUnits() const override;

private:
    OffSpecularConverter(const OffSpecularConverter& other); //!< used by clone()
    double calculateValue(size_t i_axis, Axes::Coords units, double value) const override;
    std::vector<std::map<Axes::Coords, std::string>> createNameMaps() const override;

    void addDetectorYAxis(const IDetector2D& detector);
};


//! DepthProbeConverter class handles the unit translations for depth probe simulations
//! Its default units are radians for x-axis and nm for y-axis
//! @ingroup simulation_internal

class DepthProbeConverter : public CoordSystem2D {
public:
    DepthProbeConverter() = delete;
    DepthProbeConverter(const Beam& beam, const IAxis& alpha_axis, const IAxis& z_axis);
    ~DepthProbeConverter() override;

    DepthProbeConverter* clone() const override;

    //! Returns the list of all available units
    std::vector<Axes::Coords> availableUnits() const override;

    Axes::Coords defaultUnits() const override { return Axes::Coords::DEGREES; }

private:
    DepthProbeConverter(const DepthProbeConverter& other); //!< used by clone()
    double calculateValue(size_t, Axes::Coords units, double value) const override;
    std::vector<std::map<Axes::Coords, std::string>> createNameMaps() const override;
};

#endif // BORNAGAIN_DEVICE_INSTRUMENT_COORDSYSTEM2D_H
#endif // USER_API
